@@ -15,7 +15,7 @@ require_relative 'connection_pool/timed_stack'
 #
 # Using optional timeout override (for that single invocation)
 #
-#    @pool.with(:timeout => 2.0) do |redis|
+#    @pool.with(timeout: 2.0) do |redis|
 #      redis.lpop('my-list') if redis.llen('my-list') > 0
 #    end
 #
@@ -51,6 +51,7 @@ class ConnectionPool
 
     @available = TimedStack.new(@size, &block)
     @key = :"current-#{@available.object_id}"
+    @key_count = :"current-#{@available.object_id}-count"
   end
 
 if Thread.respond_to?(:handle_interrupt)
@@ -84,20 +85,26 @@ else
 end
 
   def checkout(options = {})
-    conn = if stack.empty?
-      timeout = options[:timeout] || @timeout
-      @available.pop(timeout: timeout)
+    if ::Thread.current[@key]
+      ::Thread.current[@key_count]+= 1
+      ::Thread.current[@key]
     else
-      stack.last
+      ::Thread.current[@key_count]= 1
+      ::Thread.current[@key]= @available.pop(options[:timeout] || @timeout)
     end
-
-    stack.push conn
-    conn
   end
 
   def checkin
-    conn = pop_connection # mutates stack, must be on its own line
-    @available.push(conn) if stack.empty?
+    if ::Thread.current[@key]
+      if ::Thread.current[@key_count] == 1
+        @available.push(::Thread.current[@key])
+        ::Thread.current[@key]= nil
+      else
+        ::Thread.current[@key_count]-= 1
+      end
+    else
+      raise ConnectionPool::Error, 'no connections are checked out'
+    end
 
     nil
   end
@@ -106,25 +113,23 @@ end
     @available.shutdown(&block)
   end
 
+  # Size of this connection pool
+  def size
+    @size
+  end
+
+  # Number of pool entries available for checkout at this instant.
+  def available
+    @available.length
+  end
+
   private
 
-  def pop_connection
-    if stack.empty?
-      raise ConnectionPool::Error, 'no connections are checked out'
-    else
-      stack.pop
-    end
-  end
-
-  def stack
-    ::Thread.current[@key] ||= []
-  end
-
   class Wrapper < ::BasicObject
     METHODS = [:with, :pool_shutdown]
 
     def initialize(options = {}, &block)
-      @pool = ::ConnectionPool.new(options, &block)
+      @pool = options.fetch(:pool) { ::ConnectionPool.new(options, &block) }
     end
 
     def with(&block)
@@ -135,6 +140,14 @@ end
       @pool.shutdown(&block)
     end
 
+    def pool_size
+      @pool.size
+    end
+
+    def pool_available
+      @pool.available
+    end
+
     def respond_to?(id, *args)
       METHODS.include?(id) || with { |c| c.respond_to?(id, *args) }
     end
